function [U,Out,r_p,r_z] = ALMTVL1C(Bn,H,mu,opts)
% [U, Out] = ALMTVL1C(Bn,H,mu,opts)
%
% ALMTVL1C is a Total Variation (TV) and the 1-norm fidelity based method 
% for solving the ill-posed Grayscale image deconvolution problem of the form:
%               f = K*u + noise
% where noise is impulsive.
%
% ALMTVL1C solves 
%             min_u TV(u) + mu ||K*u - f||_1
% where TV(u) is the TV of u and is discretized as sum_i {||D_iu||}. 
%
% Inputs:
%       Bn  :  blurry and noisy observation of a grayscale image
%       H   :  convolution kernel for generating K
%       mu  :  Lagrangian multiplier
%      opts :  structure with the following fields[description | {default}]
%            r_p       - the penalty parameter for p = \nabla u
%                          [positive real  |  {10}]
%            r_z       - the penalty parameter for z = K u
%                          [positive real  |  {10}]
%            mit_inn   - maximum inner iteration number for a fixed
%                        multiplier
%                          [positive interger | {1}] 
%            U0        - initial guess of solution: [{f}]
%            issingle  - single or not? [0 or 1 | {1}]
%            idisp     - if or not display middle iteration information
%
%      Note : 1) all fields of opts are created by getFTVdopts.m
%                (see 'getFTVdopts' for more details)
%             2) local weights and higher order TV are not included 
%                for TVL1 solvers so far
%
% Outputs:
%         U : solution
%         Out : structure with the following fields
%               SNR - a vector contains history SNRs
%               Inner - Inner iteration number for each fixed beta
%               totalIter - total inner iteration number
%
% References:
%

%
% Last Modified by Juyong Zhang on Oct 12, 2009.
%

%% initialization
% get options
if nargin < 4; opts = []; end;
[r_p,r_z,rScaling,tol_inn,mit_inn,tol_outite,U,Weights,chk_opt_u,issingle,idisp] = getALMTVL1opts(opts);

if isempty(U); U = Bn; end;
[M, N, d3] = size(U); 
if d3 ~= 3
    error(['ALMTVL1C only applies to triple channel images!', ...
        'For single channel images, call ALMTVL1G']);
end

% Index
J1 = 1:M; J2 = (M+1):(2*M); J3 = (2*M+1):(3*M);
K1 = 1:N; K2 = (N+1):(2*N); K3 = (2*N+1):(3*N);

% padding zeros if H is not odd sized
% & make sure H is a 3 x 3 cell array
H = fixH(H,d3);

% compute constant quantities
[conjoDx,conjoDy,conjoKK,FS,conjotfH] = getConstantC(Bn,H);

% fix Weights
Weights = fixWeights(Weights,M,N);
Weig = Weights;
Weights = zeros(M,N,d3);
for ii = 1:d3; Weights(:,:,ii) = Weig; end;
clear Weig;

% single ?
if issingle
    conjoDx = single(conjoDx);
    conjoDy = single(conjoDy);
    conjoKK = single(conjoKK);
    FS = single(FS);
    U = single(U);
    Weights = single(Weights); 
end;

% finite diff
Ux = [diff(U,1,2)  U(:,1,:) - U(:,N,:)];
Uy = [diff(U,1,1); U(1,:,:) - U(M,:,:)];

% initialize the Lagrange multiplier lambda_p and lambda_z
lambda_p_x = zeros(M,N,d3);
lambda_p_y = zeros(M,N,d3);
lambda_z = zeros(M,N,d3);
vecNorm = sqrt(sum(Ux.*Ux+Uy.*Uy, d3)+0.001);
for ii=1:d3
    lambda_p_x(:,:,ii)=Ux(:,:,ii)./vecNorm;
    lambda_p_y(:,:,ii)=Uy(:,:,ii)./vecNorm;
end

% initialize p and Z
p_x = Ux;
p_y = Uy;
Z = imfilter33(U,H,'conv');

%
% To record some information of the procedure.
%
SNR_history = [];
Inner = [];
totalIter = 0;
Outiter = 0;
stopc = 0;

conjoDxFP1 = zeros(M,N,d3);
conjoDyFP2 = zeros(M,N,d3);
conjotfHFZ = zeros(M,N,d3);

wbar = waitbar(0, 'ALMTVL1C is running: please wait ...');

%% Main loop
while ~stopc
    Outiter = Outiter + 1;
    
    Inniter = 0;
    stopci = 0;
    while ~stopci

        % ================================
        %  Begin Alternating Minimization
        %
        % ----------------
        %   u-subprolem
        % ----------------
        for ii = 1:d3
            conjoDxFP1(:,:,ii) = conjoDx.*fft2(lambda_p_x(:,:,ii) + r_p*p_x(:,:,ii));
            conjoDyFP2(:,:,ii) = conjoDy.*fft2(lambda_p_y(:,:,ii) + r_p*p_y(:,:,ii));
        end
        FZ = fft2(lambda_z + r_z*Z);
        conjotfHFZ(:,:,1) = conjotfH(J1,K1).*FZ(:,:,1) + conjotfH(J2,K1).*FZ(:,:,2) + conjotfH(J3,K1).*FZ(:,:,3);
        conjotfHFZ(:,:,2) = conjotfH(J1,K2).*FZ(:,:,1) + conjotfH(J2,K2).*FZ(:,:,2) + conjotfH(J3,K2).*FZ(:,:,3);
        conjotfHFZ(:,:,3) = conjotfH(J1,K3).*FZ(:,:,1) + conjotfH(J2,K3).*FZ(:,:,2) + conjotfH(J3,K3).*FZ(:,:,3);
        clear FZ;
        
        Fu = getFu(conjoKK,FS,r_p,r_z,conjoDxFP1,conjoDyFP2,conjotfHFZ);
        U = real(ifft2(Fu));
        % ----------------
        %   p-subprolem
        % ----------------
        Ux = [diff(U,1,2)  U(:,1,:) - U(:,N,:)];
        Uy = [diff(U,1,1); U(1,:,:) - U(M,:,:)];
        Vx = r_p*Ux - lambda_p_x;
        Vy = r_p*Uy - lambda_p_y;
        V3 = sqrt(sum(Vx.^2 + Vy.^2,d3));
        V3(V3==0) = 1;
        V = zeros(M,N,d3);
        for ii = 1:d3; V(:,:,ii) = V3;
        end;
        V = max(1 - 1./V, 0);
        p_x = 1/r_p*V.*Vx;
        p_y = 1/r_p*V.*Vy;
        % ----------------
        %   z-subprolem
        % ----------------
        Ku = imfilter33(U,H,'conv');
        V = r_z*Ku-lambda_z;
        V = V-r_z*Bn;
        absV = abs(V);
        absV(absV==0) = mu;
        S = max(0,1-mu./absV);
        Z = Bn+1/r_z*S.*V;
        
        %
        %  End Alternating Minimization
        % ================================        
        
        Inniter = Inniter + 1;
        totalIter = totalIter + 1;
        snrU = snr(U);
        SNR_history = [SNR_history; snrU];

        % -------------------------------------------------
        % check optimality, update inner stopping criterion
        %
        %[res_u,res_wn,res_wz,zr,Ux,Uy] = checkopt_MC(Ux,Uy,U,H,Bn,beta,mu,Weights,tol_inn,chk_opt_u);     
        stopci = (Inniter >= mit_inn);
        
        %if idisp == 1
            %fprintf('  totalIter: %d, SNR %6.2fdB,',totalIter,snrU);
            %fprintf(' [res_u,res_wn,res_wz] = [%4.2e %4.2e %4.2e], ', res_u,res_wn,res_wz);
            %fprintf('Iz/mn: %4.2f\n',zr);
        %end   
    end % inner
    Inner = [Inner; Inniter];
    

    % ------------------------------------------
    % update the Lagrange multiplier and outer stopping criterion
    %
    lambda_p_x = lambda_p_x + r_p*(p_x-Ux);
    lambda_p_y = lambda_p_y + r_p*(p_y-Uy);
    lambda_z = lambda_z + r_z*(Z-Ku);
    
    err=sum(sum((p_x-Ux).^2+(p_y-Uy).^2))+sum(sum(Z-Ku).^2);
    if(err<tol_outite^2)
        stopc = 1;
    else
        stopc = 0;
    end
    
end;
close(wbar);
if ~isa(class(U), 'double')
    U = double(U);
end
Out.SNR = SNR_history;
Out.totalIter = totalIter;
Out.Inner = Inner;
%% ------------------SUBFUNCTION----------------------------------
function [conjoDx,conjoDy,conjoDD,S,conjotfH] = getConstantC(Bn,H)
%  [conjoDx,conjoDy,conjoDD,S,conjoDFBn] = getConstantC(Bn,H);
%
% Inputs: 
%        Bn -- blurry and noisy observation f
%        H  -- cross-channel convolution kernel  
% 
% Outputs:
%      conjoDx -- conjugate of eigenvalues of D^1
%      conjoDy -- conjugate of eigenvalues of D^2
%      conjoDD -- F(K'*K)
%      S       -- |F(D^1)|.^2 + |F(D^2)|.^2
%      conjotfH -- F(K')

% J. Yang, Jan. 1, 2008
% modified by Juyong Zhang on Oct 12, 2009

[M,N,P] = size(Bn);
conjoDx = conj(psf2otf([1,-1],[M,N]));
conjoDy = conj(psf2otf([1;-1],[M,N]));
S = abs(conjoDx).^2 + abs(conjoDy).^2;

if isempty(H);
    conjoDD = 1;
    return;
end;

% Index
J1 = 1:M; J2 = (M+1):(2*M); J3 = (2*M+1):(3*M);
K1 = 1:N; K2 = (N+1):(2*N); K3 = (2*N+1):(3*N);

% otfH = fft2(H)
%       = [ F(H11) F(H12) F(H13);
%           F(H21) F(H22) F(H23);
%           F(H31) F(H32) F(H33)];
otfH = zeros(3*M,3*N);
otfH(J1,K1) = psf2otf(H{1,1},[M,N]);
otfH(J1,K2) = psf2otf(H{1,2},[M,N]);
otfH(J1,K3) = psf2otf(H{1,3},[M,N]);
otfH(J2,K1) = psf2otf(H{2,1},[M,N]);
otfH(J2,K2) = psf2otf(H{2,2},[M,N]);
otfH(J2,K3) = psf2otf(H{2,3},[M,N]);
otfH(J3,K1) = psf2otf(H{3,1},[M,N]);
otfH(J3,K2) = psf2otf(H{3,2},[M,N]);
otfH(J3,K3) = psf2otf(H{3,3},[M,N]);

% conjotfH = 
%          [conj(F(H11)) conj(F(H12)) conj(F(H13));
%           conj(F(H21)) conj(F(H22)) conj(F(H23));
%           conj(F(H31)) conj(F(H32)) conj(F(H33))];
conjotfH = zeros(3*M,3*N);
conjotfH(J1,K1) = conj(otfH(J1,K1));
conjotfH(J1,K2) = conj(otfH(J1,K2));
conjotfH(J1,K3) = conj(otfH(J1,K3));
conjotfH(J2,K1) = conj(otfH(J2,K1));
conjotfH(J2,K2) = conj(otfH(J2,K2));
conjotfH(J2,K3) = conj(otfH(J2,K3));
conjotfH(J3,K1) = conj(otfH(J3,K1));
conjotfH(J3,K2) = conj(otfH(J3,K2));
conjotfH(J3,K3) = conj(otfH(J3,K3));

% F(K'*K)
conjoDD = zeros(3*M,3*N);

conjoDD(J1,K1) = conjotfH(J1,K1).*otfH(J1,K1) + conjotfH(J2,K1).*otfH(J2,K1) + conjotfH(J3,K1).*otfH(J3,K1);
conjoDD(J1,K2) = conjotfH(J1,K1).*otfH(J1,K2) + conjotfH(J2,K1).*otfH(J2,K2) + conjotfH(J3,K1).*otfH(J3,K2);
conjoDD(J1,K3) = conjotfH(J1,K1).*otfH(J1,K3) + conjotfH(J2,K1).*otfH(J2,K3) + conjotfH(J3,K1).*otfH(J3,K3);

conjoDD(J2,K1) = conjotfH(J1,K2).*otfH(J1,K1) + conjotfH(J2,K2).*otfH(J2,K1) + conjotfH(J3,K2).*otfH(J3,K1);
conjoDD(J2,K2) = conjotfH(J1,K2).*otfH(J1,K2) + conjotfH(J2,K2).*otfH(J2,K2) + conjotfH(J3,K2).*otfH(J3,K2);
conjoDD(J2,K3) = conjotfH(J1,K2).*otfH(J1,K3) + conjotfH(J2,K2).*otfH(J2,K3) + conjotfH(J3,K2).*otfH(J3,K3);

conjoDD(J3,K1) = conjotfH(J1,K3).*otfH(J1,K1) + conjotfH(J2,K3).*otfH(J2,K1) + conjotfH(J3,K3).*otfH(J3,K1);
conjoDD(J3,K2) = conjotfH(J1,K3).*otfH(J1,K2) + conjotfH(J2,K3).*otfH(J2,K2) + conjotfH(J3,K3).*otfH(J3,K2);
conjoDD(J3,K3) = conjotfH(J1,K3).*otfH(J1,K3) + conjotfH(J2,K3).*otfH(J2,K3) + conjotfH(J3,K3).*otfH(J3,K3);
 
clear otfH
%% ------------------SUBFUNCTION-----------------------------
function Fu = getFu(conjoKK,S,r_p,r_z,conjoDxFP1,conjoDyFP2,conjotfHFZ)
% This function was modified by Juyong Zhang, on Oct 12, 2009.
% Fu =
% getFu(conjoKK,S,r_p,r_z,conjoDxFP1,conjoDyFP2,conjoDxFLambdaP1,conjoDyFLambdaP2,conjotfHFZ,conjotfHFLambdaZ)
%
% This function solves
%                                        {Fur}
% {r_z * conjoKK + r_p * kron(I_3,S) }   {Fug} =
%                                        {Fub}
%                                        
% r_z * {conjotfHFZr}     {conjotfHFLambdaZr}     r_p * {conjoDxFP1r}   r_p * {conjoDyFP2r}
% r_z * {conjotfHFZg}  +  {conjotfHFLambdaZg}  +  r_p * {conjoDxFP1g} + r_p * {conjoDyFP2g} 
% r_z * {conjotfHFZb}     {conjotfHFLambdaZb}     r_p * {conjoDxFP1b}   r_p * {conjoDyFP2b}
%    {conjoDxFLambdaP1r}    {conjoDyFLambdaP2r}
% +  {conjoDxFLambdaP1g} +  {conjoDyFLambdaP2g}
%    {conjoDxFLambdaP1b}    {conjoDyFLambdaP2b}
%
%
% Inputs:
%          conjoKK -- F(K'*K)
%          S       -- |F(D^1)|.^2 + |F(D^2)|.^2
%       conjoDxFP1 -- F(Dx*Px)
%       conjoDyFP2 -- F(Dy*Py)
% conjoDxFLambdaP1 -- F(Dx*lambda_p_x)
% conjoDyFLambdaP2 -- F(Dy*lambda_p_y)
%       conjotfHFZ -- F(Dy*Z)
% conjotfHFLambdaZ -- F(Dx*lambda_z)
%
% Output: Fu -- Fourier Transform of u
%

[threeM,threeN] = size(conjoKK);

M = threeM/3;
N = threeN/3;
if (mod(M,1) ~= 0) || (mod(N,1) ~= 0)
    error('input not correct');
end

J1 = 1:M; J2 = (M+1):(2*M); J3 = (2*M+1):(3*M);
K1 = 1:N; K2 = (N+1):(2*N); K3 = (2*N+1):(3*N);

conjoKK = r_z*conjoKK;
S = r_p*S;
%
% modify the diagonal elements of conjoKK using S.
%
conjoKK(J1,K1) = conjoKK(J1,K1) + S;
conjoKK(J2,K2) = conjoKK(J2,K2) + S;
conjoKK(J3,K3) = conjoKK(J3,K3) + S;
clear S

%conjoDxFP1,conjoDyFP2,conjoDxFLambdaP1,conjoDyFLambdaP2,conjotfHFZ,conjotf
%HFLambdaZ
conjoDFBn = conjotfHFZ + (conjoDxFP1 + conjoDyFP2);

% 
% solve:
%         {Fur}   {conjoDFBn(:,:,1)}
% conjoKK {Fug} = {conjoDFBn(:,:,2)}
%         {Fub}   {conjoDFBn(:,:,3)} 
% by Gaussian elimination.
%

Tmp = conjoKK(J1,K2)./conjoKK(J1,K1);
conjoKK(J2,K2)   = conjoKK(J2,K2)   - Tmp.*conjoKK(J2,K1);
conjoKK(J3,K2)   = conjoKK(J3,K2)   - Tmp.*conjoKK(J3,K1);

Tmp = conjoKK(J1,K3)./conjoKK(J1,K1);
conjoKK(J2,K3)   = conjoKK(J2,K3)   - Tmp.*conjoKK(J2,K1);
conjoKK(J3,K3)   = conjoKK(J3,K3)   - Tmp.*conjoKK(J3,K1);

Tmp = conjoDFBn(:,:,1)./conjoKK(J1,K1);
conjoDFBn(:,:,2) = conjoDFBn(:,:,2) - Tmp.*conjoKK(J2,K1);
conjoDFBn(:,:,3) = conjoDFBn(:,:,3) - Tmp.*conjoKK(J3,K1);

Tmp = conjoKK(J3,K2)./conjoKK(J2,K2);
conjoKK(J3,K3)   = conjoKK(J3,K3)   - Tmp.*conjoKK(J2,K3);
conjoDFBn(:,:,3) = conjoDFBn(:,:,3) - Tmp.*conjoDFBn(:,:,2);

Fu = zeros(M,N,3);
Fu(:,:,3) = conjoDFBn(:,:,3)./conjoKK(J3,K3);
Fu(:,:,2) = (conjoDFBn(:,:,2) - conjoKK(J2,K3).*Fu(:,:,3))./conjoKK(J2,K2);
Fu(:,:,1) = ((conjoDFBn(:,:,1) - conjoKK(J1,K3).*Fu(:,:,3)) - conjoKK(J1,K2).*Fu(:,:,2))./conjoKK(J1,K1);