function [U,Out,r_p,r_z] = ALMTVL1G(Bn,H,mu,opts)
% [U, Out] = ALMTVL1G(Bn,H,mu,opts)
%
% ALMTVL1G is a Total Variation (TV) and the 1-norm fidelity based method 
% for solving the ill-posed Grayscale image deconvolution problem of the form:
%               f = K*u + noise
% where noise is impulsive.
%
% ALMTVL1G solves 
%             min_u TV(u) + mu ||K*u - f||_1
% where TV(u) is the TV of u and is discretized as sum_i {||D_iu||}. 
%
% Inputs:
%       Bn  :  blurry and noisy observation of a grayscale image
%       H   :  convolution kernel for generating K
%       mu  :  Lagrangian multiplier
%      opts :  structure with the following fields[description | {default}]
%            r_p       - the penalty parameter for p = \nabla u
%                          [positive real  |  {10}]
%            r_z       - the penalty parameter for z = K u
%                          [positive real  |  {10}]
%            mit_inn   - maximum inner iteration number for a fixed
%                        multiplier
%                          [positive interger | {1}] 
%            U0        - initial guess of solution: [{f}]
%            issingle  - single or not? [0 or 1 | {1}]
%            idisp     - if or not display middle iteration information
%
%      Note : 1) all fields of opts are created by getFTVdopts.m
%                (see 'getFTVdopts' for more details)
%             2) local weights and higher order TV are not included 
%                for TVL1 solvers so far
%
% Outputs:
%         U : solution
%       Out : structure with the following fields
%              SNR - a vector contains history SNRs
%            Inner - Inner iteration number for each fixed beta
%        totalIter - total inner iteration number
%
% References:
%

%
% Last Modified on Sep 23, 2009.
%

%% initialization
% get options
[r_p,r_z,rScaling,tol_inn,mit_inn,tol_outite,U,Weights,chk_opt_u,issingle,idisp] = getALMTVL1opts(opts);

if isempty(U); U = Bn; end;
[m, n, d3] = size(U); 
if d3 ~= 1
    error(['ALMTVL1G only applies to single channel images!', ...
        'For triple channel images, call ALMTVL1C']);
end

% padding zeros if H is not odd sized
% & make sure H is a 3 x 3 cell array
H = fixH(H,d3);

% get constant quantities
[conjoDx,conjoDy,conjotfH,Denom1,Denom2] = getC(Bn,H);

% single ?
if issingle
    conjoDx = single(conjoDx);
    conjoDy = single(conjoDy);
    conjotfH = single(conjotfH);
    Denom1 = single(Denom1);    
    Denom2 = single(Denom2);        
    U = single(U);
    Bn = single(Bn);
end

Denom = r_z*Denom2 + r_p*Denom1;

%
% finite diff
%
Ux = [diff(U,1,2)  U(:,1) - U(:,n)];
Uy = [diff(U,1,1); U(1,:) - U(m,:)]; 

SNR_hist = [];
Outiter = 0;
totalIter = 0;
Inner = [];
stopc = 0;

wbar = waitbar(0, 'ALMTVL1G is running, please wait ...');

%initialize the Lagrange multipliers and p and z.
lambda_p_x = zeros(size(Bn));
lambda_p_y = zeros(size(Bn));
lambda_z   = zeros(size(Bn));
%p_x = Ux;
%p_y = Uy;
%Z = imfilter(U,H,'circular','conv');

p_x = zeros(size(Bn));
p_y = zeros(size(Bn));
Z = zeros(size(Bn));

%% main loop
while ~stopc
    Outiter = Outiter + 1;
    
    stopci = 0; Inniter = 0;
    
    while ~stopci

        % ================================
        %  Begin Alternating Minimization
        % ----------------
        
        % ----------------
        %   u-subprolem
        % ----------------
        Nomin = conjotfH.*fft2(lambda_z + r_z*Z) + conjoDx.*fft2(lambda_p_x + r_p*p_x) + conjoDy.*fft2(lambda_p_y + r_p*p_y);
        FU = (Nomin)./Denom;
        U = real(ifft2(FU));

        % ----------------
        %   p-subproblem
        % ----------------
        Ux = [diff(U,1,2), U(:,1) - U(:,n)];
        Uy = [diff(U,1,1); U(1,:) - U(m,:)];
        Wx = r_p*Ux - lambda_p_x;
        Wy = r_p*Uy - lambda_p_y;
        W = sqrt(Wx.*Wx+Wy.*Wy);
        W(W==0) = 1;
        S = max(0,1-1./W);
        p_x = 1/r_p*S.*Wx;
        p_y = 1/r_p*S.*Wy;
        
        % ----------------
        %   z-subprolem
        % ----------------
        Ku = imfilter(U,H,'circular','conv');
        V = r_z*Ku-lambda_z;
        V = V-r_z*Bn;
        absV = abs(V);
        absV(absV==0) = mu;
        S = max(0,1-mu./absV);
        Z = Bn+1/r_z*S.*V;
        %
        %  End Alternating Minimization
        % ================================        
        
        snrU = snr(U);
        Inniter = Inniter + 1; totalIter = totalIter + 1;
        SNR_hist = [SNR_hist; snrU];
        
        % -------------------------------------------------
        % check optimality, update inner stopping criterion
        %
        %[res_u,res_wn,res_wz,wzr,res_Zn,res_Zz,Zzr,Ux,Uy] = ...
        %    checkopt(Wx,Wy,Z,U,H,Bn,beta,gamma,mu,tol_inn,chk_opt_u);           
        stopci = (Inniter >= mit_inn);
        
%         if idisp == 1; 
%             fprintf('  totalIter: %d, SNR %6.2fdB,',totalIter,snrU);
%             %fprintf(' [res_u,res_wn,res_wz] = [%4.2e %4.2e %4.2e], ',res_u,res_wn,res_wz);
%             %fprintf('Iz/mn: %4.2f\n',zr); 
%         end;        
    end % inner
    Inner = [Inner; Inniter];        

%     if idisp == 1
%         fprintf('******************************************');
%         fprintf('******************************************');
%         fprintf('****************** Inniter: %d\n',Inniter);
%     end

    % ------------------------------------------
    % update Lagrange multiplier and outer stopping criterion
    %
    lambda_p_x = lambda_p_x + r_p*(p_x-Ux);
    lambda_p_y = lambda_p_y + r_p*(p_y-Uy);
    lambda_z = lambda_z + r_z*(Z-Ku);
    %r = r*rScaling;
    
    err=sum(sum((p_x-Ux).^2+(p_y-Uy).^2))+sum(sum(Z-Ku).^2);
    if(err<tol_outite^2)
        stopc = 1;
    else 
        stopc = 0;
    end
    
end %Outer
close(wbar);
if ~isa(class(U), 'double')    
    U = double(U);             
end

Out.SNR = SNR_hist;
Out.totalIter = totalIter; 
Out.Inner = Inner;
%% ------------------SUBFUNCTION-----------------------------
function [conjoDx,conjoDy,conjotfH,Denom1,Denom2] = getC(Bn,H)
% [conjoDx,conjoDy,conjotfH,Denom1,Denom2] = getC(Bn,H)
%
% Inputs: 
%       Bn  -- input blurry and noisy data
%        H  -- convolution kernel  
% 
% Outputs:
%      conjoDx -- conjugate of eigenvalues of D^1
%      conjoDy -- conjugate of eigenvalues of D^2
%     conjotfH -- Fourier transform of K'
%       Denom1 -- |F(D^1)|.^2 + |F(D^2)|.^2
%       Denom2 -- |F(K)|.^2
%

% Chunlin Wu, Sep 23, 2009.

sizeB = size(Bn);

conjoDx = conj(psf2otf([1,-1],sizeB));
conjoDy = conj(psf2otf([1;-1],sizeB));
conjotfH = conj(psf2otf(H,sizeB));
Denom1 = abs(conjoDx).^2 + abs(conjoDy ).^2;
Denom2 = abs(conjotfH).^2;
%% ------------------SUBFUNCTION-----------------------------
function [res_u,res_wn,res_wz,wzr,res_Zn,res_Zz,Zzr,Ux,Uy] = ...
    checkopt(Wx,Wy,Z,U,H,Bn,beta,gamma,mu,tol_inn,chk_opt_u)      
% [res_u,res_wn,res_wz,wzr,res_Zn,res_Zz,Zzr,Ux,Uy] = ...
%   checkopt(Wx,Wy,Z,U,H,Bn,beta,gamma,mu,tol_inn,chk_opt_u)
%
% check optimality condtions of the following problem
%
%  min_{w,u} sum_i {||w_i|| + beta/2 ||w_i - D_iu||^2} 
%                           + mu ( ||z||_1 + gamma/2 ||Ku - (f + z)||^2 ).
%
% Inputs: 
%       Wx, Wy -- matrices represent {w_i: all i}
%            Z -- matrix represents z
%            U -- current iterate point
%            H -- convolution kernel
%           Bn -- blurry and noisy observation
%         beta -- penalty parameter for w
%        gamma -- penalty parameter for z
%           mu -- Lagrangian multiplier
%       tol_inn   -- inner iteration tolerance
%       chk_opt_u -- determin if check opt conds of u if all residues
%                    are smaller than tol_inn
% 
% Outputs:
%       res_u  -- residual of normal equations of u-subproblem
%       res_wn -- residual of nonzero part of W
%       res_wz -- ...         zero  ....
%       zr     -- zero ratio in W
%       res_Zn -- residual of nonzero part of Z
%       res_Zz -- ...         zero  ....
%       Zzr    -- zero ratio in Z
%       Ux, Uy -- finite differences of U
%
% For more details about optimality conditions, see
%
%   [3] J. Yang, Y. Zhang and W. Yin, "An efficient TVL1 algorithm for 
%       deblurring multichannel images corrupted by impulsive noise",
%       In processing.
%

% J. Yang, Jan. 1, 2008

[m n] = size(U);
tau = mu/sqrt(beta*gamma);
scalling = sqrt(gamma/beta);

Ux = [diff(U,1,2), U(:,1) - U(:,n)];
Uy = [diff(U,1,1); U(1,:) - U(m,:)];

Dx = Wx - Ux; 
Dy = Wy - Uy;      

V = sqrt(Wx.^2 + Wy.^2);
Iz = find(V == 0);
In = find(V ~= 0);
V(Iz) = 1;

wzr = length(Iz)/m/n;

RWx = Dx + Wx./V/beta;
RWy = Dy + Wy./V/beta;
S = sqrt(RWx.^2 + RWy.^2);
clear RWx RWy V

if isempty(In)
    res_wn = 0; 
else
    res_wn = max(S(In));
end

if res_wn > tol_inn
    res_wz = -1;
    res_Zn = 0;
    res_Zz = -1;
    Zzr = -1;
    res_u = 0;
else
    if isempty(Iz); 
        res_wz = -1;
    else
        res_wz = max(S(Iz) - 1/beta);
    end
    
    if res_wz > tol_inn
        res_Zn = 0;
        res_Zz = -1;
        Zzr = -1;
        res_u = 0;
    else
        KUmf = imfilter(U, H,'circular','conv') - Bn;
        absKUmf = abs(KUmf(:)); 
        ZIn = find( absKUmf > (tau/scalling) );
        Zzr = length(m*n - ZIn)/m/n;
        if isempty(ZIn)
            res_Zn = 0;
        else
            R3 = tau.*sign(Z(ZIn)) + scalling.*(Z(ZIn)-KUmf(ZIn));
            res_Zn = max(abs(R3));
        end
        
        if res_Zn > tol_inn
            res_Zz = -1;
            res_u = 0;
        else
            ZIz = find( absKUmf <= (tau/scalling) ); 
            if isempty(ZIz)
                res_Zz = -1;
            else
                R4 = scalling.*abs(KUmf(ZIz)) - tau;
                res_Zz = max(R4);
            end
            
            if res_Zz > tol_inn
                res_u = 0;
            elseif chk_opt_u == 0
                res_u = 0;
            else
                gK = KUmf - Z;          
                gK = imfilter(gK,H,'circular','corr');
                gK = (scalling^2)*gK;

                Dxx = [Dx(:,1)-Dx(:,n),diff(Dx,1,2)]; 
                Dyy = [Dy(1,:)-Dy(m,:);diff(Dy,1,1)];
                Ru = (Dxx + Dyy) + gK;
                res_u = max(abs(Ru(:)));
            end
        end
    end
end